<!DOCTYPE html>
<html lang="{{ site.lang | default: 'en-US' }}">

<head>
  <meta charset="UTF-8">

  {% seo %}
  <link rel="preconnect" href="https://fonts.gstatic.com">
  <link rel="preload" href="https://fonts.googleapis.com/css?family=Open+Sans:400,700&display=swap" as="style"
    type="text/css" crossorigin>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <meta name="theme-color" content="#157878">
  <meta name="apple-mobile-web-app-status-bar-style" content="black-translucent">
  {% if jekyll.environment == "production" %}
  <link rel="stylesheet"
    href="{{ site.github.repository_name | append: '/assets/css/style.css?v=' | append: site.github.build_revision | relative_url }}">
  {% else %}
  <link rel="stylesheet" href="{{ 'assets/css/style.css?v=' | append: site.github.build_revision | relative_url }}">
  {% endif %}
  {% include head-custom.html %}
</head>

<body>
  <a id="skip-to-content" href="#content">Skip to the content.</a>

  <header class="page-header" role="banner">
    <h1 class="project-name">{{ page.title | default: site.title | default: site.github.repository_name }}</h1>
    <h2 class="project-tagline">{{ page.description | default: site.description | default: site.github.project_tagline
      }}</h2>
    <a href="blueprint" class="btn">Blueprint (web)</a>
    <a href="blueprint.pdf" class="btn">Blueprint (pdf)</a>
    <a href="docs" class="btn">Documentation</a>
    {% if site.github.is_project_page %}
    <a href="{{ site.github.repository_url }}" class="btn">GitHub</a>
    {% endif %}
  </header>

  <main id="content" class="main-content" role="main">
    {{ content }}

    <footer class="site-footer">
      {% if site.github.is_project_page %}
      <span class="site-footer-owner"><a href="{{ site.github.repository_url }}">{{ site.github.repository_name }}</a>
	is maintained by {| author |}. Visit the GitHub repository for more information.</span>
      {% endif %}
    </footer>
  </main>
</body>

</html>